#include "../main/bootpack.h"
#include <stdio.h>
#include <string.h>

// 0-8̔zɂȂĂ̂̓`l0-8܂ł邩
unsigned char tmp_Op40_1[8];	// 40+OP
unsigned char tmp_Op20_0[8];	// 20+OP
unsigned char tmp_Op20_1[8];	// 20+OP
unsigned char tmp_OpB0[8];		// B0+OP

					/*     0x20          0x40          0x60          0x80          0xE0      0xC0 */
static PATCH sine  =  {{0x00, 0x21}, {0x3F, 0x00}, {0x00, 0xFF}, {0x00, 0x0F}, {0x00, 0x00}, 0x00};
static PATCH piano =  {{0x11, 0x21}, {0x90, 0x00}, {0x82, 0xF2}, {0x02, 0x73}, {0x00, 0x00}, 0x0A};
static PATCH guitar = {{0x10, 0x21}, {0xff, 0x00}, {0x82, 0xFF}, {0x02, 0x73}, {0x00, 0x00}, 0x0A};

PATCH *adlib_patch_sine(void)   { return &sine;   }
PATCH *adlib_patch_piano(void)  { return &piano;  }
PATCH *adlib_patch_guitar(void) { return &guitar; }

void io_out16(int port, int data);
void io_out8(int port, int data);
int io_in8(int port);

void adlib389w(char value);
void adlib388w(char value);
void adlibw(int regist, char value);

void beep(int hz)
{
	int i;
	if (hz == 0) {
		i = io_in8(0x61);
		io_out8(0x61, i & 0x0d);
	} else {
		i = 1193180000 / hz;
		io_out8(0x43, 0xb6);
		io_out8(0x42, i & 0xff);
		io_out8(0x42, i >> 8);
		i = io_in8(0x61);
		io_out8(0x61, (i | 0x03) & 0x0f);
	}
	return;
}

///////////////////////////////////////////////////////////////////////////////////////////
// WAV̓ǂݍ݁AĐ
///////////////////////////////////////////////////////////////////////////////////////////

#define  LONG_READ filebuf+=4
#define SHORT_READ filebuf+=2

int wav_Is_error = 0;

inline unsigned long get_ulong(char *fp)
{
	unsigned char s[4];
	memcpy(s, fp, 4);
	fp += 4;
	return (s[0] + 256L * (s[1] + 256L * (s[2] + 256L * s[3])));
}

inline unsigned get_ushort(char *fp)
{
	unsigned char s[2];
	memcpy(s, fp, 2);
	fp += 2;
	return (s[0] + 256U * s[1]);
}

struct WAV_HEADER *wave2pcm(char *filebuf, int len)
{
	char               *spt = &filebuf;
	struct WAV_HEADER  wav;

	wav_Is_error = 0;

	// RIFF wb_
	if (memcmp(filebuf, "RIFF", 4) != 0) {
		wav_Is_error = 1;
		return (struct WAV_HEADER *)-1;
	}
	filebuf += 4;
	LONG_READ;

	// WAVE wb_
	if (memcmp(filebuf, "WAVEfmt ", 8) != 0) {
		wav_Is_error = 2;
		return (struct WAV_HEADER *)-1;
	}
	filebuf += 8;
	LONG_READ;

	wav.dataformat = get_ushort(filebuf); SHORT_READ;
	wav.channels   = get_ushort(filebuf); SHORT_READ;
	wav.rate       = get_ulong (filebuf);  LONG_READ;
	LONG_READ; SHORT_READ;
	wav.bits       = get_ushort(filebuf); SHORT_READ;

	while (memcmp(filebuf, "data", 4) != 0) {filebuf++;}
	filebuf += 4;
	wav.pcmlen = get_ulong(filebuf); LONG_READ;

	wav.pcm16 = &filebuf;

	return &wav;
}

////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////

// sound.cȂ̂ɂȂVGÃ[`c?
void set_qemuvga0reg(int reg, int dat)
{
	io_out16(0x01ce, reg);
	io_out16(0x01cf, dat);
	return;
}

void init_qemuvga0(struct BOOTINFO *binfo, int x, int y, int c, int flag)
{
	set_qemuvga0reg(0x0004, 0x0000);
	set_qemuvga0reg(0x0001, x);
	set_qemuvga0reg(0x0002, y);
	set_qemuvga0reg(0x0003, c);		/* 4, 8, 15, 16, 24, 32 */
	set_qemuvga0reg(0x0005, 0x0000);
	set_qemuvga0reg(0x0004, flag);	/* jAANZX[hVRAM̏Ȃ0x41 */
		/* bit7 : VRAM}, bit6 : jAANZX[h, bit0 : ݒL */
	binfo->scrnx = x;
	binfo->scrny = y;
	binfo->vmode = c;
	if ((flag & 0x40) == 0) {
		binfo->vram = (char *) 0x000a0000;
	} else {
		binfo->vram = (char *) 0xe0000000;
	}
	return;
}

/* Adlib(FM)݂邩mFB (http://wiki.osdev.info/?AdLib) */
int found_adlib(int *fat)
{
	int i;
	char data1, data2;

	adlibw(0x04, 0x60);
	adlibw(0x04, 0x80);
	data1 = io_in8(0x388);

	adlibw(0x02, 0xff);
	adlibw(0x04, 0x21);

	/* 80}CNbȏ҂āAƂƂȂ̂Ń|[gǂ݂ĉɂԂB */
	for (i = 0; i < 125; i++)
		io_in8(0x388);

	data2 = io_in8(0x388);

	adlibw(0x04, 0x60);
	adlibw(0x04, 0x80);

	return ((data1 & 0xe0) == 0x00) && ((data2 & 0xe0) == 0xc0);
}

/* 0x388ɃANZXWX^̔ԍāA0x389ɏށB */
void adlibw(int regist, char value)
{
	adlib388w(regist);
	adlib389w(value);
	return;
}

// AdlibWX^|[g : 0x388
void adlib388w(char value)
{
	int i;
	io_out8(0x388, value);
	for (i = 0; i < 6; i++)
		io_in8(0x388);
	return;
}

// Adlibf[^|[g : 0x389
void adlib389w(char value)
{
	int i;
	io_out8(0x389, value);
	for (i = 0; i < 25; i++)
		io_in8(0x388);
	return;
}

// Iy[^ݒ
void adlibOPw(int base, int ch, int op, char value)
{
	int reg = base + ch;

	if (ch > 2)
		reg += 5;
	if (ch > 5)
		reg += 5;
	if (op)
		reg += 3;

	adlibw(reg, value);
	return;
}

// `lݒ
void adlibCHw(int base, int ch, char value)
{
	int reg = base + ch;

	adlibw(reg, value);
	return;
}

unsigned int adlibBfreq(int freq)
{
	unsigned long f_num = 2 * 524288 * freq / 50000;
	unsigned int  block = 0;

	while ((f_num > 0x3ff) && (block < 0x08))
	{
		block++;
		f_num = f_num / 2;
	}

	if (block > 7)
		return 0xffff;									// sB0xffffԂB
	else
		return (block << 10) | (unsigned int) f_num;	// I
}

// Đ
int adlib_play(int ch, int freq, int vol)
{
	unsigned int block_fnum;
	int	mult = 1;
	int	c_mult, m_mult;
    unsigned char Op20_1, Op20_2, Op40, OpA0, OpB0;		// OP

	do {
		block_fnum = adlibBfreq(freq / mult);

		// adlibBfreq() ̖߂l[0xffff]multVB
		if (block_fnum == 0xffff)
		{
			// 1A2A4A8AƂȂ悤ɂB
			mult = mult * 2;

			// mult=16ɂȂA15ɂĂ܂B
			if (mult == 16)
				mult = 15;

			// mult=15烋[v𔲂oB
			if (mult == 15)
				break;
		}
	} while (block_fnum == 0xffff);

	// calculate carrier and modulator mults
	m_mult = (tmp_Op20_0[ch] & 0x0f) * mult;
	c_mult = (tmp_Op20_1[ch] & 0x0f) * mult;

	if ((block_fnum == 0xffff) || (m_mult > 15) || (c_mult > 15))
		return 0;

	Op20_1 = (tmp_Op40_1[ch] & 0xf0) | m_mult;
	Op20_2 = (tmp_Op20_1[ch] & 0xf0) | c_mult;
	Op40   = (tmp_Op40_1[ch] & 0xc0) | (63 - vol);		// {[́u63vőŁAΏقǉ傫Ȃ̂63ňĂB

	block_fnum |= 0x2000;								// KeyONԂɂB

	OpA0  = block_fnum &  0xff;
	OpB0  = block_fnum >> 0x08;

	tmp_OpB0[ch] = OpB0;

	adlibOPw(0x20, ch, 0, Op20_1);
	adlibOPw(0x20, ch, 1, Op20_2);
	adlibOPw(0x40, ch, 1, Op40);
	adlibCHw(0xA0, ch,    OpA0);
	adlibCHw(0xB0, ch,    OpB0);

	return 1;
}

void adlib_stop(int ch)
{
	adlibCHw(0xB0, ch, tmp_OpB0[ch] & 0xdf);
	return;
}

void adlib_load_patch(int ch, PATCH *patch)
{
	adlibOPw(0x20, ch, 0, patch->chars[0]);
	adlibOPw(0x20, ch, 1, patch->chars[1]);
	adlibOPw(0x40, ch, 0, patch->ksl_lev[0]);
	adlibOPw(0x40, ch, 1, patch->ksl_lev[1]);
	adlibOPw(0x60, ch, 0, patch->att_dec[0]);
	adlibOPw(0x60, ch, 1, patch->att_dec[1]);
	adlibOPw(0x80, ch, 0, patch->sus_rel[0]);
	adlibOPw(0x80, ch, 1, patch->sus_rel[1]);
	adlibOPw(0xE0, ch, 0, patch->wav_sel[0]);
	adlibOPw(0xE0, ch, 1, patch->wav_sel[1]);
	adlibCHw(0xC0, ch,    patch->fb_conn);

	tmp_Op40_1[ch] = patch->ksl_lev[1];
	tmp_Op20_0[ch] = patch->chars[0];
	tmp_Op20_1[ch] = patch->chars[1];
	return;
}
